/*----------------------------------------------------------------------*
 * AdjList.h
 *
 * Michael Geikowski, 22.3.93
 *
 * Definiert die Klassen Node und NodeList sowie den Type AdjList zur
 * Verwaltung von Adjazenzlisten und Graphknoten
 *----------------------------------------------------------------------*/
/* Olaf Meier, 16.6.94
 *
 * Ergaenzungen:
 * ------------- 
 * Zusaetzliche Member-Routinen und Nonmember-Routinen 
 * zur Behandlung von Parsetree-Knoten und -Kanten
 * sowie die 3 neuen privaten Variablen PTminpos, PTmaxpos (Positionsangaben des
 * jeweiligen Parsetree-Unterbaumes bzgl. des Spezifikationsmoduls) und
 * PTpredecessor, eine Referenz auf den Vorgaengerknoten (NULL, falls Wurzel) 
 *-----------------------------------------------------------------------------*/ 
 
#ifndef ADJLIST_H
#define ADJLIST_H

#include "list.h"

class NodeList;

class Node : public ListElem
//==========================
	// Knoten eines Graphen
 {
 friend class NodeList;

 public:

   Node(char* name);
 //-----------------
	// Liefert neuen Knoten mit Namen name. name wird in neuen
        // Speicherbereich kopiert, zaehler werden auf Null gesetzt,
        // OutEdges auf leere Liste.

   ~Node();
 //--------
	// Loescht Name und OutEdges

   void AddEdge(Node* n, char EdgeType);
 //-------------------------------------
	// Neue Kante mit Zielknoten n und Typ EdgeType zu Liste der von this
        // ausgehenden Kanten hinzufuegen.


   void PTAddEdge(Node* n, char EdgeType);
//----------------------------------------
        // ...wie AddEdge, nur werden hier die zusaetzlichen Kanten an das Ende
        // der Liste gefuegt.

   Node* GetFirstEdge();
 //---------------------
	// Erste von this ausgehende Kante (NULL, wenn keine existiert)

   Node* GetNextEdge();
 //--------------------
	// Naechste von this ausgehende Kante (NULL, wenn keine weitere
        // existiert)


   void AddActionNr(int actionnr);

   int GetFirstActionNr();

   int GetNextActionNr();

   void ChangeCurrentEdge(Node* n);
 //--------------------------------
	// Zielknoten der aktuelle Kante des aktuellen Durchlaufs
        // (siehe GetFirstEdge, GetNextEdge) auf n aendern.

   char GetTypeOfCurrentEdge();
 //----------------------------
	// Liefert Typ der aktuellen Kante des aktuellen Durchlaufs

   char* GetName()		// Liefert Knotenname
 //---------------
	{ assert(nodeName);
	  return nodeName; }

   long GetPTminpos()  // Liefert minimale Position des Subtrees
 //-------------------
        { return PTminpos; }

   long GetPTmaxpos() // Liefert maximale Position des Subtrees
 //------------------ 
        { return PTmaxpos;}

   Node* GetPTpredecessor() // Liefert den Vorgaengerknoten (NULL, falls Wurzel)
 //------------------------
        { return PTpredecessor; }

   void SetPTpredecessor(Node* pred) 
 //---------------------------------
        { PTpredecessor = pred; }   

   void SetPTmaxpos(long mp)
 //---------------------
        { PTmaxpos = mp;}

   void SetPTminpos(long mp)
 //---------------------
        { PTminpos = mp;}

   int HasPTpredecessor();                  // Hat betrachteter Knoten einen Vorgaenger ?
 //--------------------------
       


   int HasPTsuccessor();            // Hat betrachteter Knoten (mindestens) einen Nachfolger?
 //--------------------------
       

   int GetNrOfInEdges()		// Liefert Anzahl Eingangskanten. Anzahl
 //--------------------		// wird durch obige Funktionen berechnet.
 	{ return NrOfInEdges; }

   int GetNrOfOutEdges()	// Liefert Anzahl Ausgangskanten. Anzahl
 //---------------------	// wird durch obige Funktionen berechnet.
	{ return NrOfOutEdges; }

   int GetLevel()
 //--------------
	{ return Level; }

   void SetLevel(int l)
 //--------------------
	{ Level = l; }

   int GetPos()
 //------------
	{ return Pos; }

   void SetPos(int p)
 //------------------
	{ Pos = p; }

   int GetTempPos()
 //----------------
	{ return TempPos; }

   void SetTempPos(int p)
 //----------------------
	{ TempPos = p; }

   int GetCounter()
 //----------------
	{ return Counter; }

   void SetCounter(int c)
 //----------------------
	{ Counter = c; }

   int GetCounter2()
 //-----------------
	{ return Counter2; }

   void SetCounter2(int c)
 //-----------------------
	{ Counter2 = c; }

   char GetNodeType()
 //------------------
	{ return NodeType; }

   void SetNodeType(char t)
 //------------------------
	{ NodeType = t; }




   Node* PTGoUpwards(int level);  // Geht (maximal)  'level' Kanten ausgehend vom Knoten 'this'
 //-----------------------------  // den Graph in Richtung Wurzel hinauf und gibt Zielknoten
                                  // zurueck

void Select()
 //---------------
{
  selected++;
}

void Deselect()
 //----------------
{
 selected--;
}

void Resetselected()
 {
  selected = 0;
 }

int Isselected()
 {
  return (selected > 0);
 }

 private:

   char *nodeName;	// Knotenname
   long PTminpos;       // minimale Position im des Subtrees (wird nur beim Parsetree benoetigt) 
   long PTmaxpos;       // maximale Position im des Subtrees (wird nur beim Parsetree benoetigt)
   int  NrOfInEdges;	// Anzahl ein- und
   int  NrOfOutEdges;	// ausgehender Kanten
   char NodeType;	// Knotentyp
   int  Level;		// Einige Hilfsvariablen (z.B. zur Layoutberechnung)
   int  Pos;		// Hilfsvariablen und Knotentyp koennen beliebig
   int  TempPos;	// gesetzt werden.
   int  Counter;
   int  Counter2;
   List *OutEdges;	// Liste der ausgehenden Kanten
   Node* PTpredecessor;   // Vorgaengerknoten diese Knotens (NULL ,falls Wurzel)->om 
   List *ActionList;    // Semantische Konzepterfasung : Liste der beruehrten Aktionen
   int  selected;       // zaehlt die Wahl dieses Knotens       
 };
 
class NodeList : public List
//==========================
	// Eine Liste von Knoten
 {
 public:

   NodeList()
 //----------
	{}

   ~NodeList()
 //-----------
	{}

   Node* AddNode(char* name);
 //--------------------------
 	// Teste, ob Knoten mit Name 'Name' schon in Liste existiert. Wenn ja,
        // alten Knoten zurueckliefern. Wenn nicht, neuen Knoten erzeugen und
        // zurueckliefern.

   Node* PTAddNode(char* name, char nt, long min ,long max);
 //---------------------------------------------------------
        // Teste, ob Knoten mit Name 'Name' mit Positionen min und max schon in Liste existiert.
        // Wenn ja, alten Knoten zurueckliefern. Wenn nicht, neuen Knoten erzeugen und
        // zurueckliefern. (Knoten wird hier hinten angehaengt)

   Node* SCAddNode(char* name, int actionnr);
 


   Node* GetFirstNode()
 //--------------------
	{ return (Node*) GetFirstElem(); }

   Node* GetNextNode()
 //-------------------
	{ return (Node*) GetNextElem(); }



   Node* PTSearchforNode(Node* n);
 //-------------------------------
       // Suche in der Knotenliste nach dem Knoten 'n' (verglichen wird 'name', 'PTminpos'
       // und 'PTmaxpos' und gebe diesen Knoten zurueck.


Node* PTSearchforNodeLevel(int l);


 };


 
// Typ fuer Adjazenzlisten

typedef NodeList AdjList;

// NonMember Routinen

Node* CopyAdjList(Node* start, AdjList* list, int level);

void AddNodetoAdjlist(Node* n, AdjList* l); // ... fuegt Knoten 'n' zur Adjazenzliste 'l' hinzu

void SCAddEdgesToNodes(AdjList* adjl);

#endif
