/*----------------------------------------------------------------------*
 * dbinterface.h
 *
 * Michael Geikowski, 12.5.93
 *
 * Definition der Schnittstelle zur Datenbasis.
 * Die Schnittstelle verwaltet 5 Filetypen:
 *  - GraphFile    : Modul-Hierarchie-Graphen
 *  - LinkFile     : Verweise je Modul
 *  - ContentsFile : Inhaltsverzeichnis je Modul
 *  - ErrorFile    : Fehlerliste je Modul
 *  - BookmarkFile : Lesezeichen je Spezifikation
 * Von jedem Typ kann zu einem Zeitpunkt maximal eine Datei geoeffnet
 * sein. Unterschiedliche Dateien gleichen Typs muessen unterschiedliche
 * Namen haben. Dateien unterschiedlichen Typs duerfen den gleichen
 * Namen tragen.
 * Der Praeprozessor, der ausschliesslich die Schreibfunktionen dieses
 * Moduls benutzt, kann im Standalone Modus betrieben werden. In diesem
 * Modus erzeugt er keine Informationen fuer den Browser sondern gibt
 * lediglich seine Fehlermeldungen auf stderr aus. Dieser Modus kann
 * ueber die globale Variable standalone eingeschaltet werden. Alle
 * Funktionen zum Oeffnen, Schreiben und Schliessen von Dateien bleiben
 * dann wirkungslos. Es werden lediglich Fehlermeldungen ueber
 * SimpleError() und SimpleErrorCont() auf stderr geleitet.
 *----------------------------------------------------------------------*/
/* Olaf Meier, 16.6.94
 * 
 * Ergaenzungen  
 * ------------
 * 3 zusaetzliche Filetypen:
 *  - TreeFile    : Ableitungsbaum je Modul
 *  - WarningFile : Typwarnungen je Modul
 *  - VarTreeFile  : Variablenbaum der Spezifikation
 *---------------------------------------------------------------------*/


extern int standalone;	// 1, wenn Standalone Modus, sonst 0.
extern int lineno;	// Muss beim Aufruf von SimpleError() und
			// SimpleErrorCont() die aktuelle Zeilennummer des
                        // gescannten Quelltextes enthalten.

/*----------------------------------------------------------------------*
 * OpenGraphFile
 *
 * Oeffne eine Graph-Datei mit Namen `filename' zum
 * lesen (`mode'="r") oder schreiben (`mode'="w"). Im letzten Fall wird
 * eine evtl. unter diesem Namen bereits existierende Datei geloescht.
 * width und height geben die Groesse eines zu schreibenden Graphen an.
 * Liefert True bei Erfolg, False sonst.
 *----------------------------------------------------------------------*/


int OpenGraphFile(char* filename, char* mode, int width, int height);


/*--------------------------------------------------------------------*
 * OpenVarTreeFile
 * Oeffne eine Systemvariablen-Datei mit Namen `filename' zum
 * lesen (`mode'="r") oder schreiben (`mode'="w").
 * Im letzten Fall wird
 * eine evtl. unter diesem Namen bereits existierende Datei geloescht.
 * width und height geben die Groesse eines zu schreibenden Graphen an.
 * Liefert True bei Erfolg, False sonst.
 *--------------------------------------------------------------------*/  
int OpenVarTreeFile(char* filename, char* mode, int width, int height);



/*----------------------------------------------------------------------*
 *  OpenParseTreeFile
 *
 * Oeffne eine Parsetree-Datei mit Namen `filename' zum
 * lesen (`mode'="r") oder schreiben (`mode'="w"). Im letzten Fall wird
 * eine evtl. unter diesem Namen bereits existierende Datei geloescht.
 * Liefert True bei Erfolg, False sonst.
 *----------------------------------------------------------------------*/


int OpenParseTreeFile(char* filename, char* mode);


/*----------------------------------------------------------------------*
 * GetGraphExtension
 *
 * Liefert Breite und Hoehe des aktuell gelesenen Graphen.
 *----------------------------------------------------------------------*/

 void GetGraphExtension(int *width, int *height);


/*----------------------------------------------------------------------*
 * GetVarTreeExtension
 *
 * Liefert Breite und Hoehe des aktuell gelesenen Variablenbaums.
 *----------------------------------------------------------------------*/
 void GetVarTreeExtensions(int *width, int *height);
  
/*----------------------------------------------------------------------*
 * ReadNode
 *
 * Lese einen Knoten mit Typ `t', Name `name' und Koordinaten (x,y)
 * aus aktueller Graph-Datei.
 * Alle von dem Knoten ausgehenden Kanten muessen nach dem ReadNode()
 * Aufruf durch ReadEdge() gelesen werden.
 * 'endofnodes' ist True, wenn kein Knoten mehr in der Datei enthalten
 * ist.
 * Ergibt False, bei Lesefehlern.
 *----------------------------------------------------------------------*/

 int ReadNode(char* t, char** name, int* x, int* y, int* endofnodes);


/*----------------------------------------------------------------------*
 * ReadVarNode
 *
 * Lese einen Knoten mit Typ `t', Name `name' und Koordinaten (x,y)
 * aus aktueller Systemvariablen-Datei.
 * Alle von dem Knoten ausgehenden Kanten muessen nach dem ReadNode()
 * Aufruf durch ReadEdge() gelesen werden.
 * 'endofnodes' ist True, wenn kein Knoten mehr in der Datei enthalten
 * ist.
 * Ergibt False, bei Lesefehlern.
 *----------------------------------------------------------------------*/

 int ReadVarNode(char* t, char** name, int* x, int* y, int* endofnodes); 

/*----------------------------------------------------------------------*
 * ReadPTNode
 *
 * Lese einen Knoten mit Typ `t', Name `name'
 * und Subtreepositionen (mixpos, maxpos) aus aktueller ParseTree-Graph-Datei.
 * Alle von dem Knoten ausgehenden Kanten muessen nach dem ReadPTNode()
 * Aufruf durch ReadPTEdge() gelesen werden.
 * 'endofnodes' ist True, wenn kein Knoten mehr in der Datei enthalten
 * ist.
 * Ergibt False, bei Lesefehlern.
 *----------------------------------------------------------------------*/

 int ReadPTNode(char* t,char** name, long* minpos, long* maxpos, int* endofnodes); 


/*----------------------------------------------------------------------*
 * ReadEdge
 *
 * Lese eine Kante aus der aktuellen Graph-Datei, die vom letzten 
 * gelesenen Knoten ausgeht. Die Kante hat Typ `t' und fuehrt zum 
 * Knoten `name'.
 * 'endofedges' ist True, wenn keine Kante mehr von dem Knoten ausgeht.
 * Ergibt False, bei Lesefehlern.
 *----------------------------------------------------------------------*/

 int ReadEdge(char* t, char** name, int* endofedges);

/*----------------------------------------------------------------------*
 * ReadVarEdge
 *
 * Lese eine Kante aus der aktuellen Systemvariablen-Datei, die vom letzten
 * gelesenen Knoten ausgeht. Die Kante hat Typ `t' und fuehrt zum
 * Knoten `name'.
 * 'endofedges' ist True, wenn keine Kante mehr von dem Knoten ausgeht.
 * Ergibt False, bei Lesefehlern.
 *----------------------------------------------------------------------*/

 int ReadVarEdge(char* t, char** name, int* endofedges);

/*----------------------------------------------------------------------*
 *
 * Lese die naechste Aktionsnummer des aktuellen Variablenknotens
 * aus der Systemvariablen-Datei.
 * 'endofactions' ist True, wenn der aktuelle Knoten keine weitere
 * Aktionsnummer enthaelt.
 *---------------------------------------------------------------------*/ 

 int ReadVarAction(int *action, int* endofactions);

/*----------------------------------------------------------------------*
 *  ReadPTEdge
 *
 * Lese eine Kante aus der aktuellen ParseTree-Graph-Datei, die vom letzten
 * gelesenen Knoten ausgeht. Die Kante hat Typ `t' , fuehrt zum
 * Knoten `name' und besitzt die Subtreepositionen(min,max).
 * 'endofedges' ist True, wenn keine Kante mehr von dem Knoten ausgeht.
 * Ergibt False, bei Lesefehlern.
 *----------------------------------------------------------------------*/


 int ReadPTEdge(char* t, char** name, long* min, long* max, int* endofedges); 

/*----------------------------------------------------------------------*
 * WriteNode
 *
 * Schreibe einen Knoten mit Typ `t', Name `name' und Koordinaten (x,y)
 * in aktuelle Graph-Datei.
 * Alle von dem Knoten ausgehenden Kanten muessen nach dem WriteNode()
 * Aufruf durch WriteEdge() geschrieben werden.
 *----------------------------------------------------------------------*/

 void WriteNode(char t, char* name, int x, int y);

/*----------------------------------------------------------------------*
 * WriteVarNode
 *
 * Schreibe einen Variablen-Knoten mit Typ `t', Name `name' und Koordinaten (x,y)
 * in aktuelle Systemvariablen-Datei.
 * Alle von dem Knoten ausgehenden Kanten muessen nach dem WriteVarNode()
 * Aufruf durch WriteVarEdge() geschrieben werden.
 * Die Aktionen des aktuellen Knotens muessen mit WriteVarAction() geschrieben
 * werden.
 *----------------------------------------------------------------------*/

 void WriteVarNode(char t, char* name, int x, int y);

/*----------------------------------------------------------------------*
 * WritePTNode
 *
 * Schreibe einen Knoten mit Typ `t', Name `name'
 * und Subtreepositionen (minpos, maxpos)  in aktuelle ParseTree-Graph-Datei.
 * Alle von dem Knoten ausgehenden Kanten muessen nach dem WritePTNode()
 * Aufruf durch WritePTEdge() geschrieben werden.
 *----------------------------------------------------------------------*/


 void WritePTNode(char t, char* name, long minpos, long maxpos);

/*----------------------------------------------------------------------*
 * WriteEdge
 *
 * Schreibe eine Kante in die aktuelle Graph-Datei, die vom letzten in
 * die Datei geschriebenen Knoten ausgeht. Die Kante hat Typ `t' und
 * fuehrt zum Knoten `name'.
 *----------------------------------------------------------------------*/

 void WriteEdge(char t, char* name);

/*----------------------------------------------------------------------*
 * WriteVarEdge
 *
 * Schreibe eine Kante in die aktuelle Systemvariablen-Datei, die vom letzten in
 * die Datei geschriebenen Knoten ausgeht. Die Kante hat Typ `t' und
 * fuehrt zum Knoten `name'.
 *----------------------------------------------------------------------*/

 void WriteVarEdge(char t, char* name);

/*-----------------------------------------------------------------------*
 * WriteVarAction
 * 
 * Schreibe eine Aktionsnummer in die aktuelle Systemvariablen-Datei, die 
 * zum letzten in die Datei geschriebenen Variablenknoten gehoert.
 *-----------------------------------------------------------------------*/

 void WriteVarAction(int action);


/*----------------------------------------------------------------------*
 * WritePTEdge
 *
 * Schreibe eine Kante in die aktuelle ParseTree-Graph-Datei, die vom letzten in
 * die Datei geschriebenen Knoten ausgeht. Die Kante hat Typ `t' und
 * fuehrt zum Knoten `name' und hat die Subtreepositionen (min, max).
 *----------------------------------------------------------------------*/

 void WritePTEdge(char t, char* name, long min, long max);

/*----------------------------------------------------------------------*
 * CloseGraphFile
 *
 * Schliesse die aktuelle Graph-Datei.
 *----------------------------------------------------------------------*/

void CloseGraphFile();

/*----------------------------------------------------------------------*
 * CloseVarTreeFile
 *
 * Schliesse die aktuelle Systemvariablen-Datei.
 *----------------------------------------------------------------------*/

void CloseVarTreeFile();

/*----------------------------------------------------------------------*
 * CloseParseTreeFile
 *
 * Schliesse die aktuelle ParseTree-Graph-Datei.
 *----------------------------------------------------------------------*/

void CloseParseTreeFile();

/*----------------------------------------------------------------------*
 * OpenLinkFile
 *
 * Oeffne eine Link-Datei mit Namen `filename' zum
 * lesen (`mode'="r") oder schreiben (`mode'="w"). Im letzten Fall wird
 * eine evtl. unter diesem Namen bereits existierende Datei geloescht.
 * maxpos gibt die groesste Position eines Link Startpunktes an.
 * Liefert True bei Erfolg, False sonst.
 *----------------------------------------------------------------------*/

int OpenLinkFile(char* filename, char* mode, long maxpos);

/*----------------------------------------------------------------------*
 * GetMaxLinkPos
 *
 * Wurde die Link-Datei zum Lesen geoeffnet, gibt diese Funktion die
 * maximale Startposition unter allen Verweisen der Datei aus.
 *----------------------------------------------------------------------*/

int GetMaxLinkPos();
 
/*----------------------------------------------------------------------*
 * ReadLink
 *
 * Lese einen Link aus der aktuellen Link-Datei. `pos' und `leng' geben
 * Position und Laenge des Link Startsymbols, `text' gibt Link-Text an.
 * Es folgen die Zielpunkte des Links. Speicherplatz fuer `...modname'
 * und `text' wird bereitgestellt, bzw. `*...modname' ist NULL, wenn
 * entsprechende Position gleich -1.
 * Ergibt True, wenn fehlerfrei.
 *----------------------------------------------------------------------*/

int ReadLink(long *pos,int *leng,
             char **text,
             long *defpos , char **defmodname,
             long *prevpos, char **prevmodname,
             long *nextpos, char **nextmodname,
             long *modpos , char **destmodname);

/*----------------------------------------------------------------------*
 * EndOfLinkFile()
 *
 * Ergibt True, wenn Link-Datei zum Lesen geoeffnet wurde und alle
 * Links durch ReadLink gelesen wurden.
 *----------------------------------------------------------------------*/

int EndOfLinkFile();

/*----------------------------------------------------------------------*
 * WriteLink
 *
 * Schreibe einen neuen Link in die aktuelle Link-Datei. `pos' und `leng'
 * geben Position und Laenge des Symbols, das Ausgangspunkt des Links ist.
 * `s1', `s2' und `s3' werden konkatentiert und ergeben den Link-Text.
 * Es folgen die Zielpunkte des Links. Wird ein Zielpunkt nicht
 * genutzt, muss sein Name gleich NULL sein.
 *----------------------------------------------------------------------*/

void WriteLink(long pos,int leng,
               char* s1, char* s2, char* s3,
               long defpos , char* defmodname,
               long prevpos, char* prevmodname,
               long nextpos, char* nextmodname,
               long modpos , char* destmodname);

/*----------------------------------------------------------------------*
 * CloseLinkFile
 *
 * Schliesse die aktuelle Link-Datei.
 *----------------------------------------------------------------------*/

void CloseLinkFile();

/*----------------------------------------------------------------------*
 * OpenErrorFile
 *
 * Oeffne eine Error-Datei mit Namen `filename' zum
 * lesen (`mode'="r") oder schreiben (`mode'="w"). Im letzten Fall wird
 * eine evtl. unter diesem Namen bereits existierende Datei geloescht.
 * Liefert True bei Erfolg, False sonst.
 *----------------------------------------------------------------------*/

int OpenErrorFile(char* filename, char* mode);

/*----------------------------------------------------------------------*
 * OpenWarningFile
 *
 * Oeffne eine Warning-Datei mit Namen `filename' zum
 * lesen (`mode'="r") oder schreiben (`mode'="w"). Im letzten Fall wird
 * eine evtl. unter diesem Namen bereits existierende Datei geloescht.
 * Liefert True bei Erfolg, False sonst.
 *----------------------------------------------------------------------*/

int OpenWarningFile(char* filename, char* mode);

/*----------------------------------------------------------------------*
 * GetNrOfErrors
 *
 * Wurde die Error-Datei zum Lesen geoeffnet, gibt diese Funktion die
 * Anzahl der Zeilen, die in der Datei enthalten sind, aus.
 * (Eine Fehlermeldung kann sich ueber mehrere Zeilen erstrecken)
 *----------------------------------------------------------------------*/

int GetNrOfErrors();

/*----------------------------------------------------------------------*
 * GetNrOfWarnings
 *
 * Wurde die Warningr-Datei zum Lesen geoeffnet, gibt diese Funktion die
 * Anzahl der Zeilen, die in der Datei enthalten sind, aus.
 * (Eine Warnungsmeldung kann sich ueber mehrere Zeilen erstrecken)
 *----------------------------------------------------------------------*/


int GetNrOfWarnings();
 
/*----------------------------------------------------------------------*
 * SimpleError
 *
 * Schreibe eine Fehlermeldung `msg' (ohne Newline) in aktuelle Fehler-
 * Datei mit Fehlerort `errpos', `errleng'. Soll die Meldung mehrzeilig
 * werden, kann sie mit SimpleErrorCont() fortgesetzt werden.
 * Ueber die globale Variable `standalone' kann das Schreiben in die Datei
 * verhindert werden. Statt dessen wird die Meldung auf stderr ausgegeben.
 *----------------------------------------------------------------------*/

void SimpleError(char *msg, long errpos, int errleng);


/*----------------------------------------------------------------------*
 * SimpleErrorCont
 *
 * Setzt die letzte Fehlermeldung fort ( siehe SimpleError() ) um die
 * Zeichenkette `msg'. Das Ende einer Zeile wird durch Newline
 * gekennzeichnet. Die letzte Zeile wird beim naechsten Aufruf von
 * SimpleError() oder CloseErrorFile() automatisch abgeschlossen.
 * Ueber die globale Variable `standalone' kann das Schreiben in die Datei
 * verhindert werden. Statt dessen wird die Meldung auf stderr ausgegeben.
 *----------------------------------------------------------------------*/

void SimpleErrorCont(char *msg);

/*----------------------------------------------------------------------*
 * SimpleWarning
 *
 * Schreibe eine Warnung `msg' (ohne Newline) in aktuelle Warnungs-
 * Datei mit Ort `errpos', `errleng'. Soll die Meldung mehrzeilig
 * werden, kann sie mit SimpleWarningCont() fortgesetzt werden.
 * Ueber die globale Variable `standalone' kann das Schreiben in die Datei
 * verhindert werden. Statt dessen wird die Meldung auf stderr ausgegeben.
 *----------------------------------------------------------------------*/


void SimpleWarning(char *msg, long warnpos, int warnleng, int line);

/*----------------------------------------------------------------------*
 * SimpleWarningCont
 *
 * Setzt die letzte Warnungsmeldung fort ( siehe SimpleWarning() ) um die
 * Zeichenkette `msg'. Das Ende einer Zeile wird durch Newline
 * gekennzeichnet. Die letzte Zeile wird beim naechsten Aufruf von
 * SimpleWarning() oder CloseWarningFile() automatisch abgeschlossen.
 * Ueber die globale Variable `standalone' kann das Schreiben in die Datei
 * verhindert werden. Statt dessen wird die Meldung auf stderr ausgegeben.
 *----------------------------------------------------------------------*/

void SimpleWarningCont(char *msg);




/*----------------------------------------------------------------------*
 * ReadErrorLine
 *
 * Lese eine Zeile `line' aus der Fehler-Datei mit Fehlerort
 * `pos', `leng'.
 * Ergibt False bei Lesefehler.
 *----------------------------------------------------------------------*/

int ReadErrorLine(char** line, long* pos, int* leng);

/*----------------------------------------------------------------------*
 * ReadWarningLine
 *
 * Lese eine Zeile `line' aus der Warnungs-Datei mit Ort der Warnung
 * `pos', `leng'.
 * Ergibt False bei Lesefehler.
 *----------------------------------------------------------------------*/


int ReadWarningLine(char** line, long *pos, int* leng);

/*----------------------------------------------------------------------*
 * CloseErrorFile
 *
 * Schliesse die aktuelle Error-Datei.
 * Gibt die Anzahl geschriebener Fehlermeldungen zurueck (nur bei
 * mode = "w").
 *----------------------------------------------------------------------*/

int CloseErrorFile();


/*----------------------------------------------------------------------*
 * CloseWarningFile
 *
 * Schliesse die aktuelle Warnungs-Datei.
 * Gibt die Anzahl geschriebener Warnungsmeldungen zurueck (nur bei
 * mode = "w").
 *----------------------------------------------------------------------*/

int CloseWarningFile();

/*----------------------------------------------------------------------*
 * OpenContentsFile
 *
 * Oeffne eine Inhaltsverzeichnis-Datei mit Namen `filename' zum
 * lesen (`mode'="r") oder schreiben (`mode'="w"). Im letzten Fall wird
 * eine evtl. unter diesem Namen bereits existierende Datei geloescht.
 * Liefert True bei Erfolg, False sonst.
 *----------------------------------------------------------------------*/

int OpenContentsFile(char* filename, char* mode);

/*----------------------------------------------------------------------*
 * GetNrOfContentsLines
 *
 * Wurde die Inhalts-Datei zum Lesen geoeffnet, gibt diese Funktion die
 * Anzahl der Zeilen des gespeicherten Inhaltsverzeichnisses aus.
 *----------------------------------------------------------------------*/

int GetNrOfContentsLines();
 
/*----------------------------------------------------------------------*
 * BeginContentsLine
 *
 * Schreibe eine Zeile in die Inhaltsverzeichnis-Datei. Die Zeile wird
 * `depth' weit eingerueckt, stueckweise mit AddToContentsLine()
 * geschrieben und mit EndContentsLine() abgeschlossen.
 *----------------------------------------------------------------------*/

void BeginContentsLine(int depth);

/*----------------------------------------------------------------------*
 * AddToContentsLine
 *
 * Fuege eine weitere Zeichenkette (ohne Space und Newline) zur aktuellen
 * Zeile hinzu. Die Zeichenketten werden automatisch durch Space getrennt.
 * ( siehe BeginContentsLine() )
 *----------------------------------------------------------------------*/

void AddToContentsLine(char *s);

/*----------------------------------------------------------------------*
 * EndContentsLine
 *
 * Schliesse aktuelle Zeile ab mit Sprungziel `pos', `leng'.
 * ( siehe BeginContentsLine() )
 *----------------------------------------------------------------------*/

void EndContentsLine(long pos, int leng);

/*----------------------------------------------------------------------*
 * WriteContentsLine
 *
 * Schreibe eine Zeile in die Inhaltsverzeichnis-Datei mit Sprungziel
 * `pos', `leng'. Die Zeile ergibt sich aus Konkatenation von `s1' und
 * `s2' (`s2' kann auch NULL sein) und wird `depth' weit eingerueckt.
 * ( Alternative zu obigen drei Funktionen )
 *----------------------------------------------------------------------*/

void WriteContentsLine(int depth, char* s1, char* s2, long pos, int leng);

/*----------------------------------------------------------------------*
 * ReadContentsLine
 *
 * Lese eine Zeile `line' aus der Inhaltsverzeichnis-Datei mit Sprungziel
 * `pos', `leng'.
 * Ergibt False bei Lesefehler.
 *----------------------------------------------------------------------*/

int ReadContentsLine(char** line, long* pos, int* leng);

/*----------------------------------------------------------------------*
 * CloseContentsFile
 *
 * Schliesse die aktuelle Inhaltsverzeichnis-Datei.
 *----------------------------------------------------------------------*/

void CloseContentsFile();

/*----------------------------------------------------------------------*
 * OpenBookmarkFile
 *
 * Oeffne eine Lesezeichen-Datei mit Namen `filename' zum
 * lesen (`mode'="r") oder schreiben (`mode'="w"). Im letzten Fall wird
 * eine evtl. unter diesem Namen bereits existierende Datei geloescht.
 * Liefert True bei Erfolg, False sonst.
 *----------------------------------------------------------------------*/

int OpenBookmarkFile(char* filename, char* mode);

/*----------------------------------------------------------------------*
 * GetNrOfBookmarks
 *
 * Wurde die Lesezeichen-Datei zum Lesen geoeffnet, gibt diese Funktion die
 * Anzahl der gespeicherten Lesezeichen aus.
 *----------------------------------------------------------------------*/

int GetNrOfBookmarks();
 
/*----------------------------------------------------------------------*
 * WriteBookmark
 *
 * Schreibe ein Lesezeichen in die Lesezeichen-Datei mit Sprungziel
 * `ModName', `pos' und Name 'name'.
 *----------------------------------------------------------------------*/

void WriteBookmark(char* name, char* ModName, long pos);

/*----------------------------------------------------------------------*
 * ReadBookmark
 *
 * Lese ein Lesezeichen aus der Lesezeichen-Datei mit Sprungziel
 * `ModName', `pos' und Name 'name'.
 * Ergibt False bei Lesefehler.
 *----------------------------------------------------------------------*/

int ReadBookmark(char** name, char** ModName, long* pos);

/*----------------------------------------------------------------------*
 * CloseBookmarkFile
 *
 * Schliesse die aktuelle Lesezeichen-Datei.
 *----------------------------------------------------------------------*/

void CloseBookmarkFile();

