/*----------------------------------------------------------------------*
 * piLayout.c
 *
 * Olaf Meier, 16.08.94
 *
 * Erstellt Layout des Process/Interface-Graphen
 *----------------------------------------------------------------------*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "pilayout.h"
#include "pqueuecxx.h"


// Einige das Layout beeinflussende Konstante
//-------------------------------------------

#define MAXUPDOWN	10 /*2*/	// Anzahl Durchlaeufe von Phase 2
#define LAYOUTVSKIP     4	// Vertikaler Abstand zwischen 2 Level
#define LAYOUTHSKIP     4       // Minimaler horizontaler Abstand zwischen
				// zwei benachbarten Knoten eines Levels
				// (sollte immer gerade sein)

// ALIAS - Namen fuer einige AdjList Funktionen (werden in Phase 3 verwendet)
//---------------------------------------------------------------------------

#define GETXPOS()             GetTempPos()
#define SETXPOS(x)            SetTempPos(x)
#define GETRIGHTSPACE()       GetCounter2()
#define SETRIGHTSPACE(r)      SetCounter2(r)
#define GETXPOSFIX()          GetLevel()
#define SETXPOSFIX(f)         SetLevel(f)

// Modulglobale Typen
//-------------------

enum version   {oldPos, newPos};
enum direction {upBarry, downBarry};

// Modulglobale Variablen
//-----------------------

static int NewNameCounter;	// Wird benoetigt zur Erzeugung
				// eindeutiger Namen von Dummy-Knoten
static int NrOfLevels;		// Enthaelt nach phase1a die
				// Zahl der Level des Graphen
static int maxpos;		// Graphbreite

/*----------------------------------------------------------------------*
 *                            P H A S E   1
 *----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*
 * phase1a
 *
 * Einteilung der Knoten in Level (erstes Level hat Nr. 0).
 * Ergibt True, wenn Graph azyklisch, sonst False.
 * Seiteneffekt: bestimmt `NrOfLevels' als Anzahl der Level des Graphen
 *----------------------------------------------------------------------*/

static char phase1a(AdjList *al)
 {
 int  SeenAllNodes;
 int  l,NrOfNodesInGraph;
 Node *n,*m;

 NrOfNodesInGraph = al->GetLength();
 
 l = -1;				// aktueller Level
 n = al->GetFirstNode();		// `Counter' auf Anzahl der
 while (n)				// Vorgaenger initialisieren
   {
   n->SetCounter(n->GetNrOfInEdges());
   n = al->GetNextNode();
   }
 do					// Leveleinteilung
   {
   l++;
   SeenAllNodes = 1;
   n = al->GetFirstNode();		// Knoten, deren Vorgaenger alle bereits
   while (n)				// eingeteilt wurden, kommen ins
     {					// aktuelle Level
     if (n->GetCounter() == 0)
       {
       SeenAllNodes = 0;
       n->SetCounter(-1);		// Knoten merken
       n->SetLevel(l);
       }
     n = al->GetNextNode();
     }
   n = al->GetFirstNode();		// Fuer alle Knoten Anzahl der bereits
   while (n)				// eingeteilten Knoten korregieren
     {
     if (n->GetCounter() == -1)		// wenn Knoten gemerkt
       {
       n->SetCounter(-2);		// Knoten vergessen
       NrOfNodesInGraph--;
       m = n->GetFirstEdge();
       while (m)
         {
         m->SetCounter(m->GetCounter()-1); // Ein Vorgaenger weniger
         m = n->GetNextEdge();
         }
       }
     n = al->GetNextNode();
     }
   } while ( ! SeenAllNodes );		// Bis alle Knoten vergessen (oder
   					// bis Zyklus gefunden)
 NrOfLevels = l;
 return (NrOfNodesInGraph==0);		// Graph zyklenfrei, wenn alle Knoten
 					// eingeteilt wurden.
 }

/*----------------------------------------------------------------------*
 * NewDummyNode
 *
 * Erzeugt einen Dummy-Knoten 'Node' auf dem Level 'Level' mit einem
 * neuen eindeutigen Namen.
 *----------------------------------------------------------------------*/

static Node* NewDummyNode(AdjList *al, int Level)
 {
 Node *n;
 char NewName[10];

 NewNameCounter++;			// Zaehler zur Generierung der Namen
 sprintf(NewName,"#%d",NewNameCounter);
 n = al->AddNode(NewName);
 n->SetNodeType('D');			// Typ ist dummy Knoten
 n->SetLevel(Level);
 return (n);
 }

/*----------------------------------------------------------------------*
 * phase1b
 *
 * Lange Kanten segmentieren
 *----------------------------------------------------------------------*/

static void phase1b(AdjList *al)
 {
 Node *n,*m,*p,*q;
 int  nLevel,mLevel,i;
 char t;

 n = al->GetFirstNode();		// Fuer jeden Knoten n
 while (n)
   {
   nLevel = n->GetLevel();
   m = n->GetFirstEdge();		// Fuer alle Nachfolger m von n
   while (m)
     {
     mLevel = m->GetLevel();
     if (mLevel != nLevel+1)		// Wenn lange Kante
       {
       p = NewDummyNode(al,nLevel+1);	// Erzeuge von n bis m einen Pfad mit
       t = n->GetTypeOfCurrentEdge();	// je einem dummy Knoten auf Level
       n->ChangeCurrentEdge(p);		// nLevel+1 bis mLevel-1 
       for (i=nLevel+2; i<mLevel; i++)
         {
         q = NewDummyNode(al,i);
         p->AddEdge(q,t);
         p = q;
         }
       p->AddEdge(m,t);
       } 
     m = n->GetNextEdge();
     }
   n = al->GetNextNode();
   }
 }

/*----------------------------------------------------------------------*
 *                            P H A S E   2
 *----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*
 * InitCounter
 *
 * Setze die Counter der Knoten auf Level 'Level' von 'hg' auf Null.
 *----------------------------------------------------------------------*/

static void InitCounter(HiGraph *hg, int Level)
 {
 Node *n;

 n = hg->FirstNodeOfLevel(Level);
 while (n)
   {
   n->SetCounter(0);
   n->SetCounter2(0);
   n = hg->NextNodeOfLevel(Level);
   }
 }

/*----------------------------------------------------------------------*
 * SumPosOfPreds
 *
 * Summiert in Counter die Positionen aller direkten Vorgaenger je
 * Knoten auf Level 'Level' von 'hg'.
 * Position : 'version == old' : Pos, 'version == new' : TempPos. 
 *----------------------------------------------------------------------*/

static void SumPosOfPreds(HiGraph *hg, int Level, enum version ver)
 {
 Node *n,*m;
 int pos;

 if (Level>0)
   {
   n = hg->FirstNodeOfLevel(Level-1);	// Je Knoten n des Vorgaengerlevels
   while (n)
     {
     if (ver == oldPos)
       {
       pos = n->GetPos();
       }
     else
       {
       pos = n->GetTempPos();
       }
     m = n->GetFirstEdge();		// Addiere Position von n zu allen
     while (m)				// Nachfolgern m von n
       {
       m->SetCounter(m->GetCounter()+pos);
       m = n->GetNextEdge();
       }
     n = hg->NextNodeOfLevel(Level-1);
     }
   }
 }

/*----------------------------------------------------------------------*
 * SumPosOfSuccs
 *
 * Summiert in Counter2 die Positionen aller direkten Nachfolger je
 * Knoten auf Level 'Level' von 'hg'.
 * Position : 'version == old' : Pos, 'version == new' : TempPos. 
 *----------------------------------------------------------------------*/

static void SumPosOfSuccs(HiGraph *hg, int Level, enum version ver)
 {
 Node *n,*m;

 n = hg->FirstNodeOfLevel(Level);	// Je Knoten n des Levels
 while (n)
   {
   m = n->GetFirstEdge();		// Addiere die Positionen aller
   while (m)				// Nachfolger m von n
     {
     if (ver == oldPos)
       {
       n->SetCounter2(n->GetCounter2()+m->GetPos());
       }
     else
       {
       n->SetCounter2(n->GetCounter2()+m->GetTempPos());
       }
     m = n->GetNextEdge();
     }
   n = hg->NextNodeOfLevel(Level);
   }
 }

/*----------------------------------------------------------------------*
 * CountCrossings
 *
 * Liefert Anzahl der Kantenueberschneidungen zwischen Level 'Level' und
 * 'Level' + 1 von `hg' bzgl. der aktuellen ('v' == oldPos) oder
 * der temporaeren ('v' == newPos) Positionen der Knoten.
 *----------------------------------------------------------------------*/

static int CountCrossings(HiGraph *hg, int Level, enum version v)
 {
 Node *n,*m;
 Node **NodeArray;
 int  i,j,cross,pos;

 NodeArray = new Node*[hg->GetNodesOnLevel(Level)]; // Knoten des Levels `Level'
 n = hg->FirstNodeOfLevel(Level);	// gemaess gewuenschter Position in
 while (n)				// ein Array schreiben
   {
   if (v == newPos)
     {
     NodeArray[n->GetTempPos()-1] = n;
     }
   else
     {
     NodeArray[n->GetPos()-1] = n;
     }
   n = hg->NextNodeOfLevel(Level);
   }
 // Jede Kante e, die von einem Knoten v mit Position i des Levels `Level'
 // ausgeht, wird verglichen mit jeder Kante f, die von einem Knoten w mit
 // Position j>i des Levels `Level' ausgeht. Ist die Position des Zielknotens
 // von j groesser als die Position des Zielknotens von i, so ueberschneiden
 // sich die Kanten. Auf diese Weise werden alle Ueberschneidungen gefunden.
 cross = 0;
 for (i=0; i < hg->GetNodesOnLevel(Level)-1; i++)	// Fuer alle v
   {
   n = NodeArray[i]->GetFirstEdge();			// Fuer alle e
   while (n)
     {
     pos = n->GetPos();
     for (j=i+1; j < hg->GetNodesOnLevel(Level); j++) 	// Fuer alle w
       {
       m = NodeArray[j]->GetFirstEdge();		// Fuer alle f
       while (m)
         {
         if (m->GetPos() < pos)
           {
           cross++;					// Ueberschneidung
           }
         m = NodeArray[j]->GetNextEdge();
         }
       }
     n = NodeArray[i]->GetNextEdge();
     }
   }
 delete [] NodeArray;
 return (cross);
 }

/*----------------------------------------------------------------------*
 * ImprovePos
 *
 * Setze die Position der Knoten auf Level 'Level' von 'hg' gleich
 * ihrer temporaeren Position, falls dadurch die Anzahl der Kanten-
 * ueberschneidungen zwischen Level 'Level-1' und 'Level' ('dir == 
 * upBarry') bzw. 'Level' und 'Level+1' ('dir == downBarry') abnimmt.
 *----------------------------------------------------------------------*/

static void ImprovePos(HiGraph *hg, int Level, enum direction dir)
 {
 Node *n;
 int  l;

 if (dir == upBarry)
   l = Level-1;
 else
   l = Level;
 if ( CountCrossings(hg,l,newPos) <= CountCrossings(hg,l,oldPos) )
   {
   n = hg->FirstNodeOfLevel(Level);
   while (n)
     {
     n->SetPos(n->GetTempPos());
     n = hg->NextNodeOfLevel(Level);
     }
   }
 }

/*----------------------------------------------------------------------*
 * ImproveLevel
 *
 * Versucht die Positionen der Knoten auf Level 'Level' von 'HiGraph'
 * neu zu sortieren, so dass die Anzahl der Kantenueberschneidungen
 * des Graphen abnimmt. Grundlage der Verbesserung ist das Up-BarryCenter
 * ('dir == upBarry') oder das Down-BarryCenter ('dir == downBarry').
 *----------------------------------------------------------------------*/

static void ImproveLevel(HiGraph *hg, int Level, enum direction dir)
 {
 PQueue *queue;
 Node   *n;
 int    pos;
 float  prio;

 // Initialisierung

 InitCounter(hg,Level);
 if (dir == downBarry)
   {
   SumPosOfSuccs(hg,Level,oldPos);
   }
 SumPosOfPreds(hg,Level,oldPos);
 queue = hg->GetQueueOfLevel(Level);
 queue->Empty();
 
 // Knoten (heuristisch) mit Prioritaeten versehen und in Priority Queue
 // schreiben.
 // Prioritaet bei upBarry: SPv / ( Ae * AKv ) + alte Position/1000
 // mit SPv := Summe der Positionen der direkten Vorgaengerknoten
 //     Ae  := Anzahl eingehende Kanten
 //     AKv := Anzahl Knoten auf Vorgaengerlevel
 // Prioritaet bei downBarry: SPn / ( Aa * AKn ) + alte Position/1000
 // mit SPn := Summe der Positionen der direkten Nachfolgerknoten
 //     Aa  := Anzahl ausgehende Kanten
 //     AKn := Anzahl Knoten auf Nachfolgerlevel
 // (Bei Knoten ohne ausgehende Kanten (Aa=0) siehe upBarry)
 // Spv/Ae bzw. SPn/Aa sollen dafuer sorgen, dass ein Knoten moeglichst in der
 // Mitte unter seinen Vorgaengern, bzw. ueber seinen Nachfolgern steht. Das
 // Teilen soll die Prioritaeten bezueglich upBarry und downBarry vergleichbar
 // machen fuer den Fall Aa=0, wo upBarry und downBarry Prioritaeten
 // gemischt werden. Die Addition von `alte Position/1000' sorgt dafuer, dass
 // Knoten mit gleicher Prioritaet in gleicher Reihenfolge bleiben.
 
 n = hg->FirstNodeOfLevel(Level);
 while (n)
   {
   if ((dir == upBarry) || (n->GetNrOfOutEdges() == 0))
     {
     prio = ((float) n->GetCounter()) / 
            ((float) n->GetNrOfInEdges() * hg->GetNodesOnLevel(Level-1));
     }
   else
     {
     prio = ((float) n->GetCounter2()) /
            ((float) n->GetNrOfOutEdges() * hg->GetNodesOnLevel(Level+1));
     }
   queue->Insert(n,prio + ((float) n->GetPos()) / 1000.0);
   n = hg->NextNodeOfLevel(Level);
   }
   
 // Temporaere Positionen gemaess Prioritaeten verteilen. Queue liefert Knoten
 // mit hoechster Prioritaet zuerst.
 
 pos = hg->GetNodesOnLevel(Level);
 while (! queue->IsEmpty())
   {
   n = queue->DeleteMax();
   n->SetTempPos(pos);
   pos--;
   }
   
 // Vergleiche temporaere Reihenfolge mit alter und aendere Reihenfolge
 // gegebenenfalls
 
 ImprovePos(hg,Level,dir);
 }

/*----------------------------------------------------------------------*
 * phase2
 *
 * Reihenfolge der Knoten je Level festlegen.
 *----------------------------------------------------------------------*/

#ifdef LAYOUT_TEST

static void out2(HiGraph *hg)
	// Ausgabe fuer Test-Faelle
 {
 Node *n;
 int  i,j,ok;

 printf("\n---------------------------------------\n");
 for (i=0; i<NrOfLevels; i++)
   {
   j = 1;
   ok = 1;
   while (ok)
     {
     n = hg->FirstNodeOfLevel(i);
     ok = 0;
     while (n)
       {
       if ( n->GetPos() == j )
         {
         ok = 1;
         printf("%s ",n->GetName());
         break;
         }
       n = hg->NextNodeOfLevel(i);
       }
     j++;
     }
   printf("\n(%d)\n\n",CountCrossings(hg,i,oldPos));
   }
 }
#endif

static void phase2(HiGraph *hg)
 {
 int Level,updown;

 // ??? Hier vielleicht erst dafuer sorgen, dass keine dummy-Knoten mehr
 // links aussen stehen (Tausche Position mit erstem nicht-dummy):
 //
 //      a                                  a
 //     /|\                                /|\
 //    < b c	wuerde dann so aussehen:  b | c
 //     \|/                                \|/
 //      d                                  d

 for (updown = 1; updown <= MAXUPDOWN; updown++)  // Laufe MAXUPDOWN mal den
   {						  // Graphen rauf und runter
   for (Level = NrOfLevels-2; Level > 0; Level--)
     {
     ImproveLevel(hg,Level,downBarry);	// Versuche rauflaufend zu verbessern
     }
#ifdef LAYOUT_TEST
 out2(hg);
#endif
   for (Level=1; Level < NrOfLevels; Level++)
     {
     ImproveLevel(hg,Level,upBarry);	// Versuche runterlaufend zu verbessern
     }
#ifdef LAYOUT_TEST
 out2(hg);
#endif
   }
 }

/*----------------------------------------------------------------------*
 *                            P H A S E  3
 *----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*
 * NodeWidth
 *
 * Liefert Breite eines Knoten 'n'.
 *----------------------------------------------------------------------*/

static int NodeWidth(Node *n)
 { 
 char *name;

 name = n->GetName();
 
 if (name[0] == '#')
   {
   return(1+LAYOUTHSKIP);		// dummy Knoten
   }
 else
   {
   return(strlen(name)+LAYOUTHSKIP);	// echter Knoten
   }
 }

/*----------------------------------------------------------------------*
 * ShiftRight
 *
 * Schiebe den Knoten mit Index 'nr' von 'NodeArray' (und evtl. einige
 * rechts von ihm liegende) so weit wie moeglich (max. 'distance'
 * Einheiten) nach rechts. 'maxnr' gibt die Anzahl der Knoten in
 * 'NodeArray' an.
 * RIGHTSPACE enthaelt fuer jeden Knoten den Abstand zu seinem rechten
 * Nachbarn.
 * Knoten mit XPOSFIX < 0 duerfen mit Ausnahme des Knoten mit Index `nr'
 * nicht mehr bewegt werden.
 *----------------------------------------------------------------------*/

static void ShiftRight(Node* NodeArray[], int nr, int maxnr, int distance)
 {
 int i,delta;

 // bestimme in delta die Anzahl der Einheiten, um die der Knoten nach
 // rechts verschoben werden kann.

 delta = NodeArray[nr]->GETRIGHTSPACE();
 for (i=nr+1; (i<maxnr) && 
              (NodeArray[i]->GETXPOSFIX()>=0) &&
              (delta < distance)             ; i++)
   {
   delta += NodeArray[i]->GETRIGHTSPACE(); // Summiere verfuegbare Freiraeume
   }
 if ( i == maxnr )	// rechts vom letzten Knoten ist beliebig Platz
   {
   delta = distance;
   }
 maxnr = i;		// Knoten von nr bis maxnr-1 werden verschoben
 if (nr > 0)		// evtl. rechten Freiraum von linkem Nachbarn von
   {			// Knoten nr korregieren
   NodeArray[nr-1]->SETRIGHTSPACE(NodeArray[nr-1]->GETRIGHTSPACE()+delta);
   }
 for (i=nr; i<maxnr; i++)
   {
   NodeArray[i]->SETXPOS(NodeArray[i]->GETXPOS()+delta);
   delta -= NodeArray[i]->GETRIGHTSPACE();
   NodeArray[i]->SETRIGHTSPACE(0);   
   }
 }

/*----------------------------------------------------------------------*
 * ShiftLeft
 *
 * Schiebe den Knoten mit Index 'nr' von 'NodeArray' (und evtl. einige
 * links von ihm liegende) so weit wie moeglich (max. 'distance'
 * Einheiten) nach links.
 * RIGHTSPACE enthaelt fuer jeden Knoten den Abstand zu seinem rechten
 * Nachbarn.
 * Knoten mit XPOSFIX < 0 duerfen mit Ausnahme des Knoten mit Index `nr'
 * nicht mehr bewegt werden.
 *----------------------------------------------------------------------*/

static void ShiftLeft(Node* NodeArray[], int nr, int distance)
 {
 int i,delta,minnr;

 // bestimme in delta die Anzahl der Einheiten, um die der Knoten nach
 // links verschoben werden kann.

 delta = 0;
 for (i=nr-1; (i>=0) && 
              (NodeArray[i]->GETXPOSFIX()>=0) &&
              (delta < distance)             ; i--)
   {
   delta += NodeArray[i]->GETRIGHTSPACE(); // Summiere verfuegbare Freiraeume
   }
 if ( i == (-1) )	// links vom ersten Knoten ist beliebig Platz
   {
   delta = distance;
   }
 else if ( NodeArray[i]->GETXPOSFIX() < 0 )
   {
   delta += NodeArray[i]->GETRIGHTSPACE(); // Hier ist noch etwas Platz
   NodeArray[i]->SETRIGHTSPACE(0);
   }
 if (delta > distance)	// vielleicht zuviel Platz gesammelt, dann korregieren
   {
   NodeArray[i]->SETRIGHTSPACE(delta-distance);
   delta = distance;
   }
 minnr = i;		// Knoten von minnr+1 bis nr werden verschoben
 NodeArray[nr]->SETRIGHTSPACE(NodeArray[nr]->GETRIGHTSPACE()+delta);
 NodeArray[nr]->SETXPOS(NodeArray[nr]->GETXPOS()-delta);
 for (i=nr-1; i>minnr; i--)
   {
   delta -= NodeArray[i]->GETRIGHTSPACE();
   NodeArray[i]->SETXPOS(NodeArray[i]->GETXPOS()-delta);
   NodeArray[i]->SETRIGHTSPACE(0);
   }
 }

/*----------------------------------------------------------------------*
 * SumPosOfSuccsOfNode
 *
 * Liefert die Summe der Positionen aller direkten Nachfolger des
 * Knotens n.
 *----------------------------------------------------------------------*/

static int SumPosOfSuccsOfNode(Node *n)
 {
 Node *m;
 
 int sum = 0;
 for (m = n->GetFirstEdge(); m!=0; m = n->GetNextEdge())
   {
   sum += m->GETXPOS();
   }
 return sum;
 }

/*----------------------------------------------------------------------*
 * phase3
 *
 * Koordinaten der Knoten je Level festlegen.
 * Die Reihenfolge der Knoten je Level bleibt unangetastet. Ziel der
 * Phase 3 ist es, den Knoten Koordinaten in einem zweidimensionalen
 * Koordinatensystem zuzuordnen. Die Koordinatenangaben sollen aus
 * Paaren (x,y) natuerlicher Zahlen x und y bestehen, die sich auf den
 * Mittelpunkt des Knotens beziehen. Ein Knoten hat die Hoehe 1. Die
 * Breite entspricht der Laenge seines Namens plus einer Konstante
 * LAYOUTHSKIP.
 * Die y-Position eines Knotens auf Level l wird als (l+1)*LAYOUTVSKIP
 * angenommen. Fuer die Berechnung der x-Position werden die Knoten je
 * Level zunaechst so angeordnet, dass der Mittelpunkt des i-ten Knoten an
 * x-Position j liegt, wobei j die Summe der Breiten der Knoten 1..(i-1)
 * ist. Durch horizontales Verschieben der Knoten (wie Perlen auf einer
 * Schnur), d.h. durch Einfuegen von Freiraeumen zwischen den Knoten bzw.
 * vor dem ersten Knoten, wird (heuristisch) versucht eine optisch
 * ansprechende Anordnung der Knoten der einzelnen Level zueinander zu
 * erreichen.
 * Die Knoten erhalten voruebergehend eventuell negative x-Positionen
 * zugewiesen, was am Ende der Phase 3 korregiert wird.
 *----------------------------------------------------------------------*/

static int phase3(HiGraph *hg)
 {
 Node   *n;
 Node   **NodeArray;
 PQueue *queue;
 int    Level,i,bestpos,pos,minpos,width,maxlevelwidth,widestlevel;
 char*  name;

 // Initialisiere Level 0

 n = hg->FirstNodeOfLevel(0);
 n->SETXPOS(0);			// x-Position
 n->SETRIGHTSPACE(0);		// Freiraum zum rechten Nachbarn

 // Einmal den Graphen von Level 1 bis Level NrOfLevels-1 durchlaufen

 maxlevelwidth = 1;		// Anzahl Knoten und
 widestlevel   = 0;		// Nr des Levels mit den meisten Knoten
 for (Level=1; Level < NrOfLevels; Level++)
   {
   if (hg->GetNodesOnLevel(Level) > maxlevelwidth) // widestlevel aktualisieren
     {
     maxlevelwidth = hg->GetNodesOnLevel(Level);
     widestlevel   = Level;
     }
   NodeArray = new Node* [hg->GetNodesOnLevel(Level)];

   queue = hg->GetQueueOfLevel(Level);
   queue->Empty();
   n = hg->FirstNodeOfLevel(Level);	// Knoten gemaess Reihenfolge in ein
   while (n)				// Array eintragen sowie mit
     {					// Prioritaet in Priority Queue
     NodeArray[n->GetPos()-1] = n;

     name = n->GetName();
     if (name[0] == '#')		// Prioritaet:
       {
       queue->Insert(n,10000.0);	// dummy Knoten 10000
       }
     else				// echte Knoten Anzahl Eingangskanten
       {
       queue->Insert(n,(float) n->GetNrOfInEdges());
       }
     n = hg->NextNodeOfLevel(Level);
     }
   pos = 0;				// Initiale x-Positionen
   for (i=0; i<hg->GetNodesOnLevel(Level); i++)
     {
      width = NodeWidth(NodeArray[i]); 
      NodeArray[i]->SETXPOS(pos+width/2); 
      pos += width;
     }
   InitCounter(hg,Level);		// Freiraum zu rechtem Nachbarn gleich 0
   SumPosOfPreds(hg,Level,newPos);	// Summe der x-Positionen der direkten
   					// Vorgaengerknoten nach Counter
   for (i=0; i<hg->GetNodesOnLevel(Level); i++)
     {
     n = queue->DeleteMax();		// Knoten gemaess Prioritaet betrachten
     n->SETXPOSFIX(-1);			// x-Position fixieren
     bestpos = n->GetCounter() / n->GetNrOfInEdges(); // Beste Position:
     					// moeglichst zentral unter Vorgaenger
     if (n->GETXPOS() < bestpos)	// Schiebe so nah wie moeglich an
       {				// beste Position heran
       ShiftRight(NodeArray,n->GetPos()-1,hg->GetNodesOnLevel(Level),
                  bestpos-n->GETXPOS());
       }
     else if (n->GETXPOS() > bestpos)
       {
       ShiftLeft(NodeArray,n->GetPos()-1,n->GETXPOS()-bestpos);
       }
     }
   delete [] NodeArray;
   }

 // Laufe von Level mit den meisten Knoten zu Level 0 zurueck. Aus
 //
 // a                        a
 // |\                      / \
 // b c     wird dadurch    b  c
 // | |\                    | / \
 // d e f                   d e f
 
 for (Level=widestlevel-1; Level >= 0; Level--)
   {
   NodeArray = new Node* [hg->GetNodesOnLevel(Level)];
  
   queue = hg->GetQueueOfLevel(Level);
   queue->Empty();
   n = hg->FirstNodeOfLevel(Level);
   while (n)
     {
     NodeArray[n->GetPos()-1] = n;

     name = n->GetName();
     if ( (name[0] != '#') && (n->GetNrOfOutEdges()>0) )
       {
       // Echte Knoten mit mind. einer Ausgangskante sollen evtl. noch einmal
       // verschoben werden und zwar (siehe Prioritaetsberechnung) zuerst die
       // aeusseren Knoten des Levels und zum Schluss der innerste Knoten.
       queue->Insert(n,
         (float) abs(n->GetPos()-1-hg->GetNodesOnLevel(Level)/2) );
       // ??? Vieleicht andere Reihenfolge oder Phase 3 up iterieren
       // ( 3 Knoten rechts, dritter soll nach links, geht nicht wegen Mitte )
       }
     n = hg->NextNodeOfLevel(Level);
     }
   while ( !queue->IsEmpty() )
     {
     n = queue->DeleteMax();
     bestpos = SumPosOfSuccsOfNode(n) / n->GetNrOfOutEdges();
     // Beste Position moeglichst zentral ueber Nachfolger
     if (n->GETXPOS() < bestpos)
       {
       ShiftRight(NodeArray,n->GetPos()-1,hg->GetNodesOnLevel(Level),
                  bestpos-n->GETXPOS());
       }
     else if (n->GETXPOS() > bestpos)
       {
       ShiftLeft(NodeArray,n->GetPos()-1,n->GETXPOS()-bestpos);
       }
     }
   delete [] NodeArray;
   }

 n = hg->FirstNodeOfLevel(0);		// negative Positionen korregieren
 width=NodeWidth(n);
 minpos = n->GETXPOS()-width/2;		// kleinste Position unter allen
 					// Positionen der linken Knotenraender
 maxpos = minpos + width;		// groesste Position unter allen
 					// Positionen der rechten Knotenraender
 for (Level=1; Level < NrOfLevels; Level++)
   {
   n = hg->FirstNodeOfLevel(Level);
   while (n)
     {
     width=NodeWidth(n);
     pos = n->GETXPOS()-width/2;	// linker Rand
     if (pos < minpos) minpos = pos;
     pos = pos + width;			// rechter Rand
     if (pos > maxpos) maxpos = pos;
     n = hg->NextNodeOfLevel(Level);
     }
   }
 maxpos = maxpos - minpos;		// maxpos = Graphbreite
 for (Level=0; Level < NrOfLevels; Level++)
   {
   n = hg->FirstNodeOfLevel(Level);
   while (n)
     {
     n->SETXPOS(n->GETXPOS()-minpos);	// kleinste Position unter allen
     					// Positionen der linken Knotenraender
                                        // wird Null
     n = hg->NextNodeOfLevel(Level);
     }
   }
 }

#ifdef LAYOUT_TEST

static void out3(HiGraph *hg)
	// Ausgabe fuer Testzwecke
 {
 Node *n;
 int  i,j,k,ok,width;
 char* name;

 printf("\n---------------------------------------\n");
 for (i=0; i<NrOfLevels; i++)
   {
   j = 1;
   ok = 1;
   while (ok)
     {
     n = hg->FirstNodeOfLevel(i);
     ok = 0;
     while (n)
       {
       if ( n->GetPos() == j )
         {
         if (j == 1)
           {
           width = NodeWidth(n);
           for (k=0; k<n->GETXPOS()-(width/2); k++) printf(" ");
           }
         ok = 1;
         name = n->GetName();
         if (name[0]=='#')
           {
           printf(" # ");
           }
         else
           {
           printf(" %s ",name);
           }
         for (k=0; k<n->GETRIGHTSPACE(); k++) printf(" ");
         break;
         }
       n = hg->NextNodeOfLevel(i);
       }
     j++;
     }
   printf("\n(%d)\n\n",CountCrossings(hg,i,oldPos));
   }
 }

#endif



     

/*----------------------------------------------------------------------*
 *                            E X P O R T E
 *----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*
 * PILayout
 *
 * Hauptprozedur: Startet die Bearbeitung der einzelnen Phasen 
 * 
 *----------------------------------------------------------------------*/



void PILayout(AdjList *al,int *PIwidth, int *PIheight)
 {
   assert(al);
   Node    *n,*m;
   int     Level,y,acyclic;



    // PHASE 1: Leveleinteilung
   acyclic = phase1a(al);

   if (! acyclic) return;
  
   phase1b(al);

  // PHASE 2: Permutation der Knoten je Level mit moeglichst wenigen
  //          Kantenueberschneidungen

   HiGraph *hg = new HiGraph(al,NrOfLevels);
   phase2(hg);


 // PHASE 3: Koordinaten fuer Knoten berechnen

   phase3(hg);

  //PHASE 4: Koordinaten je Knoten abspeichern

   y = 0;
   for(Level=0;Level<NrOfLevels;Level++)
    {
      y+= LAYOUTVSKIP;
      n = hg->FirstNodeOfLevel(Level);
      while(n)
      {
       n->SetPos(n->GETXPOS());
       n->SETXPOS(y);
       n = hg->NextNodeOfLevel(Level);
      }
    }


 *PIwidth = maxpos;

 *PIheight = (NrOfLevels+1)*(LAYOUTVSKIP);

  delete hg;
 } 
