/*-------------------------------------------------------------------------
 * Typeckeck.h
 *
 * Olaf Meier,  10.5.94
 *
 * Dieses Modul uebernimmt die Typpruefung in den einzelnen Modulen der Spezifikation.
 * Die Hauptroutine TypeChecker() startet die Pruefung auf Grundlage des jeweiligen
 * Ableitungsbaums.
 * Die Namen der benutzten Routinen sind an den jeweiligen Produktionsregeln der
 * tTLA+ Grammatik angelehnt. Der Ableitungsbaum wird durch die Routinen traversiert.
 *-------------------------------------------------------------------------------------*/    



#ifndef TYPECHECK_H
#define TYPECHECK_H
#include "parsetree.h"
#include "symlist.h"
#include "tlatype.h"




// Hauptroutine der Typpruefung

void TypeChecker(ParseExpr* tree);


// Typvergleich der Typkonstrukte 'typ1' und 'typ2'. Bei den jeweiligen
// Konstrukten handelt es sich entweder um eine Verkettung von meheren Typen oder
// eine Parametertypenliste. 

int CompareTypeChain(TypeComp* typ1, TypeComp* typ2, int isElemOf = 0);


// Typvergleich der Typkonstrukte 'typ1' und 'typ2'.

int CompareTypes(TypeComp* typ1, TypeComp* typ2, int isElemOf = 0);


// Behandlung der Produktionsregel 'Include_Parts'.
// Ggf. werden hier Parametertypen substituiert.

void Handle_IncludeParts(ParseExpr* expr);

// Behandlung der Produktionsregel 'Include_Part'.
// Eine Liste aller includierten Bezeichner wird zurueckgegeben.
// Ruft die Routine Handle_IncludePart() auf.

SymList*  Handle_IncludePart(ParseExpr* expr);

// Typaktualisierung von globalen Parametern des Moduls, die zwar
// definiert , jedoch im betreffenden Modul nicht angewendet werden
// (Typen globaler Parameter werden erst bei ihrer ersten Anwendung
// ermittelt).
 
void MakeActualTypeInfos(SymList* list);

// Behandlung der Produktionsregel 'SubstitutionPart01'
// Die Bezeichnerliste 'sl' wird zur Parametersubstitution
// durchlaufen.

void Handle_SubstitutionPart01(ParseExpr* expr,SymList* sl);

// Behandlung der Produktionsregel 'Substitution'
// Die Bezeichnerliste 'sl' wird zur Parametersubstitution
// durchlaufen. Diese Routine ruft Handle_SubstitutionPart01()  auf.

void Handle_Substitution(ParseExpr* expr,SymList* l);


// Substitution von Bezeichnertyp  'par' durch den Typ 'typ'.
// Die Typsubstitutionen erfolgt in den Typangaben der Bezeichner 
// aus der Liste 'inclist'.
   
void ReplacePar(SymListElem* par, TypeComp* typ,SymList* inclist);

// Behandlung der Produktionsregel 'expr_list'.
void Handle_expr_list(ParseExpr* expr);

// Behandlung der Produktionsregeln aller Ausdruecke im Definitions-
// abschnitt CONSTANTS

void Handle_ConstantDefinitions(ParseExpr* expr,noTerminal_Type nt);

// Behandlung der Produktionsregeln aller Formel-Definitionsabschnitte

void Handle_BoolDefinitions(ParseExpr* expr,noTerminal_Type nt);

// Behandlung aller restlichen Definitionsabschnitte.
void Handle_RestDefinitions(ParseExpr* expr,noTerminal_Type nt);


// Behandlung der Produktionsregel 'SymbolDef_LHS'.
// Flag 'nt' gibt die Bezeichnung des jeweiligen Definitionsabschnitts an.
// Es wird der neu definierte Bezeichner zurueckgegeben.

SymListElem* AppendType_SymbolDef_LHS(ParseExpr* expr, noTerminal_Type nt);


// Ermittlung des Tupelselektors des Terminalsymbols 'TUPELSELECTOR' 
// Der jeweilige Selektorzahl wird zurueckgegeben.

int GetSelector(SymListElem* selector);


// Behandlung der Produktionsregel 'expr_csl'.
// Der Typausdruck 'prev' enthaelt die Typenverkettung
// der Ausdruecke nach Ausfuehrung der Routine.
// Der jeweilige Aufruf erfolgt rekursiv.
// Schliesslich wird die vollstandig aufgebaute Typenverkettung
// zurueckgegeben. 
 
TypeComp* GetType_expr_csl(ParseExpr* expr,TypeComp  *prev = NULL);

// Behandlung der Produktionsregel 'qualifierPrefixes'.
// Besitzt ein Qualifier Modulparameter, so werden diese durch die Typen der
// aktuellen Argumente ersetzt.
// Der aktuelle Typ des betrachteten Identidiers wird zurueckgegeben.

TypeComp* CheckType_qualifierPrefixes(ParseExpr* expr, TypeComp* idtyp);

// Traegt das Typkonstrukt 'tp' bei Bezeichner 'se' ein.

void AddTypeInfo(SymListElem* se, TypeComp* tp);


// Ermittlung des Typs vom Teilbaum 'expr'.
// Der entsprechende Typausdruck wird zurueckgegeben.

TypeComp* CreateType(ParseExpr* expr );

// Ermittlung des Typs eines angewendeten Bezeichners, der im
// Teilbaum 'expr' dargestellt wird.
// Der entsprehende Typausdruck wird zurueckgeliefert.

TypeComp* GetTypeofID(ParseExpr* expr);

// Behandlung der Produktionsregel 'rctype_ssl'. (Rekordkonstrukt)
// Der entsprechende Typausdruck wird zurueckgeliefert.

TypeComp* AppendRecComp_rctype_ssl(ParseExpr* expr,TypeComp* prev = NULL);

// Behandlung der Produktionsregel 'rctype'. (Rekordkonstrukt)
// Der entsprechende Typausdruck wird zurueckgeliefert.

TypeComp* AppendRecComp_rctype(ParseExpr* expr, TypeComp* prev = NULL);

// Verbindung einzelner Rekordkomponenten
// Die Komponenten erhalten den jeweils den Typausdruck 'comptype'.

TypeComp* Connect_RecComps(ParseExpr* expr,TypeComp* comptype, TypeComp* prev = NULL);

// Verbindung einzelner Rekordkomponenten der Produktionsregel 'recdef_csl'.
// Die entsprechende Komponentenverketung wird zurueckgegeben.

TypeComp* AppendRecComp_recdef_csl(ParseExpr* expr,TypeComp* prev = NULL);


// Verbindung einzelner Rekordkomponenten der Produktionsregel 'recdef'.
// Die entsprechende Komponentenverketung wird zurueckgegeben.

TypeComp* AppendRecComp_recdef(ParseExpr* expr, TypeComp* prev = NULL);

// Behandlung der Produktionsregel 'isElementOfSetListQB'.
// Handelt es sich um ein ARRAY-Ausdruck ('isFunc' == 1), werden die vorkommenden Variablen
// als ARRAY-Parameter behandelt.
// Der entsprechende Typausdruck wird zurueckgegeben. 

TypeComp* AppendTypes_isElementOfSetListQB(ParseExpr* expr,int isFunc = 0,  TypeComp* prev = NULL);

// Behandlung der Produktionsregel 'normalID_csl'.
// Flag 'isFuncOrOp' gibt an, ob die Bezeichner innerhalb eines ARRAY- bzw. OPERATOR-
// Konstrukts als Parameter benutzt werden (==1).
// Der entsprechende Typausdruck wird zurueckgegeben.

TypeComp* SetType_normalID_csl(ParseExpr* expr,TypeComp* tp,int isFuncOrOp  = 0,
                                                                   TypeComp* prev = NULL);

// Behandlung der Produktionsregel 'formalPar'.
// Flag 'isFuncOrOp' gibt an, ob 'formalPar' ein  ARRAY- bzw. OPERATOR-
// Konstrukt darstellt (==1).
// Der entsprechende Typausdruck wird zurueckgegeben.

TypeComp* AppendType_formalPar(ParseExpr* expr,int isFuncOrOp= 0, TypeComp* prev = NULL);


// Behandlung der Produktionsregel 'formalPar_ssl'.
// Flag 'isFuncOrOp' gibt an, ob  'formalPar_ssl' formale Parameter
// eines ARRAY- bzw. OPERATOR-Konstrukts darstellen (==1).
// Der entsprechend aufgebaute Typausdruck wird zurueckgegeben.

TypeComp* AppendType_formalPar_ssl(ParseExpr* expr, int isFuncOrOp =  0,TypeComp* prev = NULL);


// Behandlung der Produktionsregel 'prefixLines'.
// Der entsprechende Typausdruck wird zurueckgegeben.

TypeComp* GetType_prefixLines(ParseExpr* expr);

// Behandlung der Produktionsregel 'con_dis_Lines'.
// Der entsprechende Typausdruck wird zurueckgegeben.

TypeComp* GetType_con_dis_Lines(ParseExpr* expr);

// Behandlung der Produktionsregel 'con_dis_Lines'.
// Der entsprechende Typausdruck wird zurueckgegeben.
// Routine ruft GetType_CaseRA() auf.

TypeComp* GetType_CaseRA_List(ParseExpr* expr);


// Behandlung der Produktionsregel 'CaseRA'.
// Der entsprechende Typausdruck wird zurueckgegeben.

TypeComp* GetType_CaseRA(ParseExpr* expr);

// Behandlung der Produktionsregel 'SetExpr'.
// Der entsprechend aufgebaute Typausdruck wird zurueckgegeben.

TypeComp* ComposeType_SetExpr(ParseExpr* expr);

// Behandlung der Produktionsregel 'expr'.
// Der entsprechend aufgebaute Typausdruck wird zurueckgegeben.

TypeComp* ComposeType_expr(ParseExpr* expr);


#endif
